use std::sync::{Arc, Mutex};
use std::time::{Duration, Instant};

/// this is just a big dumb bag of perf-related metrics,
/// gathered with the intention of identifying areas for optimization
pub struct Diagnostics {
    frames_completed: u64,
    time_spent_stepping: Duration,
    time_spent_on_stimuli: Duration,
    time_spent_prepping_render: Duration,
    time_spent_blocked_on_render: Duration,
    time_spent_rendering: Duration,
    start_time: Instant,
}

#[derive(Clone, Default)]
pub struct SyncDiagnostics(Arc<Mutex<Diagnostics>>);

impl Default for Diagnostics {
    fn default() -> Self {
        Self::new()
    }
}

impl Diagnostics {
    pub fn new() -> Self {
        Self {
            frames_completed: 0,
            time_spent_stepping: Default::default(),
            time_spent_on_stimuli: Default::default(),
            time_spent_prepping_render: Default::default(),
            time_spent_blocked_on_render: Default::default(),
            time_spent_rendering: Default::default(),
            start_time: Instant::now(),
        }
    }

    pub fn format(&self) -> String {
        let step_time = self.time_spent_stepping.as_secs_f64();
        let stim_time = self.time_spent_on_stimuli.as_secs_f64();
        let render_time = self.time_spent_rendering.as_secs_f64();
        let render_prep_time = self.time_spent_prepping_render.as_secs_f64();
        let block_time = self.time_spent_blocked_on_render.as_secs_f64();
        let overall_time = self.start_time.elapsed().as_secs_f64();
        let fps = (self.frames_completed as f64) / overall_time;
        format!("fps: {:6.2} (sim: {:.1}s, stim: {:.1}s, [render: {:.1}s], blocked: {:.1}s, render_prep: {:.1}s, other: {:.1}s)",
            fps,
            step_time,
            stim_time,
            render_time,
            block_time,
            render_prep_time,
            overall_time - step_time /*- stim_time*/ - block_time - render_prep_time,
        )
    }
}

impl SyncDiagnostics {
    pub fn new() -> Self {
        Self(Arc::new(Mutex::new(Diagnostics::new())))
    }
    pub fn format(&self) -> String {
        self.0.lock().unwrap().format()
    }
    /// measure the duration of some arbitrary chunk of code.
    /// used internally.
    pub fn measure<R, F: FnOnce() -> R>(f: F) -> (Duration, R) {
        let start = Instant::now();
        let r = f();
        (start.elapsed(), r)
    }

    /// record the duration of the sim step operation.
    pub fn instrument_step<R, F: FnOnce() -> R>(&self, frames: u64, f: F) -> R {
        let (elapsed, ret) = Self::measure(f);
        let mut me = self.0.lock().unwrap();
        me.time_spent_stepping += elapsed;
        me.frames_completed += frames as u64;
        ret
    }

    /// record the duration spent preparing for render (i.e. cloning stuff and moving it into a
    /// render pool).
    pub fn instrument_render_prep<R, F: FnOnce() -> R>(&self, f: F) -> R {
        let (elapsed, ret) = Self::measure(f);
        self.0.lock().unwrap().time_spent_prepping_render += elapsed;
        ret
    }

    /// record the duration actually spent doing CPU render work
    pub fn instrument_render_cpu_side<R, F: FnOnce() -> R>(&self, f: F) -> R {
        let (elapsed, ret) = Self::measure(f);
        self.0.lock().unwrap().time_spent_rendering += elapsed;
        ret
    }

    /// record the duration spent blocking the simulation because the render queue is full.
    pub fn instrument_render_blocked<R, F: FnOnce() -> R>(&self, f: F) -> R{
        let (elapsed, ret) = Self::measure(f);
        self.0.lock().unwrap().time_spent_blocked_on_render += elapsed;
        ret
    }

    pub fn instrument_stimuli<R, F: FnOnce() -> R>(&self, f: F) -> R {
        let (elapsed, ret) = Self::measure(f);
        self.0.lock().unwrap().time_spent_on_stimuli += elapsed;
        ret
    }

}

