use ndarray::{self, Array3};
use serde::{Deserialize, Serialize};
use log::{info, trace, warn};

use crate::diagnostics::SyncDiagnostics;
use crate::geom::{Coord, Index};
use crate::real::Real;
use crate::sim::{AbstractSim, Fields, GenericSim, StaticSim};
use crate::stim::AbstractStimulus;
use crate::cross::vec::Vec3;
use coremem_cross::mat::{FullyGenericMaterial, Material};
use coremem_cross::step::SimMeta;

mod cpu;
mod gpu;
pub use cpu::CpuBackend;
pub use gpu::WgpuBackend;

/// while the SpirvSim type owns all the stateful data,
/// it dispatches to a pluggable backend for advancing that state.
///
/// this backend type abstracts over the specific compute used to advance the state.
/// it could advance the state locally, on the CPU,
/// or it could dispatch to a SPIR-V device, like a GPU.
pub trait SimBackend<R, M> {
    fn step_n(
        &mut self,
        meta: SimMeta<R>,
        mat: &[M],
        stim_e: &[Vec3<R>],
        stim_h: &[Vec3<R>],
        e: &mut [Vec3<R>],
        h: &mut [Vec3<R>],
        m: &mut [Vec3<R>],
        num_steps: u32,
    );
}

/// Wrapper around an inner state object which offloads stepping onto a spirv backend (e.g. GPU).
#[derive(Default, Serialize, Deserialize)]
// TODO: remove default R/M
pub struct SpirvSim<R=f32, M=FullyGenericMaterial<R>, B=WgpuBackend>
where M: 'static
{
    meta: SimMeta<R>,
    e: Vec<Vec3<R>>,
    h: Vec<Vec3<R>>,
    m: Vec<Vec3<R>>,
    mat: Vec<M>,
    step_no: u64,
    #[serde(skip)]
    backend: B,
    #[serde(skip)]
    diag: SyncDiagnostics,
}

// B isn't always clonable (e.g. gpu backend) so rust can't auto-derive this.
impl<R: Clone, M: Clone, B: Default> Clone for SpirvSim<R, M, B> {
    fn clone(&self) -> Self {
        Self {
            meta: self.meta.clone(),
            e: self.e.clone(),
            h: self.h.clone(),
            m: self.m.clone(),
            mat: self.mat.clone(),
            step_no: self.step_no.clone(),
            // backends can be expensive to clone.
            // we require that the caller explicitly init the backend if they need this.
            // TODO: this probably shouldn't be a `Clone` method, but like "read_only_clone()".
            backend: Default::default(),
            diag: self.diag.clone(),
        }
    }
}

impl<R, M, B> AbstractSim for SpirvSim<R, M, B>
where
    R: Real,
    M: Send + Sync + Material<R> + Clone + Into<FullyGenericMaterial<R>>,
    B: Send + Sync + SimBackend<R, M>,
{
    type Real = R;
    type Material = M;

    fn meta(&self) -> SimMeta<f32> {
        self.meta.cast()
    }
    fn step_no(&self) -> u64 {
        self.step_no
    }

    fn put_material_index(&mut self, pos: Index, mat: M) {
        let flat_idx = self.flat_index(pos).unwrap();
        self.mat[flat_idx] = mat;
    }

    fn get_material_index(&self, pos: Index) -> &M {
        let flat_idx = self.flat_index(pos).unwrap();
        &self.mat[flat_idx]
    }

    fn step_multiple<S: AbstractStimulus>(&mut self, num_steps: u32, stim: &S) {
        let (stim_e, stim_h) = self.eval_stimulus(stim);
        self.backend.step_n(
            self.meta,
            self.mat.as_slice(),
            stim_e.as_slice().unwrap(),
            stim_h.as_slice().unwrap(),
            self.e.as_mut_slice(),
            self.h.as_mut_slice(),
            self.m.as_mut_slice(),
            num_steps,
        );
        self.step_no += num_steps as u64;
    }

    fn fields_at_index(&self, pos: Index) -> Fields<R> {
        let flat_idx = self.flat_index(pos);
        match flat_idx {
            Some(idx) => Fields::new(
                self.e[idx], self.h[idx], self.m[idx]
            ),
            None => Default::default(),
        }
    }

    fn to_static(&self) -> StaticSim {
        let mut mat = Vec::new();
        mat.resize(self.e.len(), Default::default());
        StaticSim {
            meta: self.meta.cast(),
            e: self.e.iter().map(|v| v.cast()).collect(),
            h: self.h.iter().map(|v| v.cast()).collect(),
            m: self.m.iter().map(|v| v.cast()).collect(),
            mat,
            step_no: self.step_no,
            backend: Default::default(),
            diag: self.diag.clone(),
        }
    }

    fn to_generic(&self) -> GenericSim<Self::Real> {
        let mat = self.mat.iter().map(|m| m.clone().into()).collect();
        GenericSim {
            meta: self.meta,
            e: self.e.clone(),
            h: self.h.clone(),
            m: self.m.clone(),
            mat,
            step_no: self.step_no,
            backend: Default::default(),
            diag: self.diag.clone(),
        }
    }

    fn use_diagnostics(&mut self, diag: SyncDiagnostics) {
        self.diag = diag;
    }
}

impl<R: Real, M: Default, B> SpirvSim<R, M, B>
{
    pub fn new_with_backend(size: Index, feature_size: f32, backend: B) -> Self {
        info!("SpirvSim::new({:?}, {})", size, feature_size);
        let flat_size = size.volume() as usize;
        if flat_size * std::mem::size_of::<M>() >= 0x40000000 {
            warn!("simulation size ({:?}) is greater than what is tested for: anything could happen!", size);
        }
        let feature_size = feature_size.cast::<R>();
        let courant = 0.577.cast::<R>();
        let time_step = feature_size / R::c() * courant;
        let mut mat = Vec::new();
        mat.resize_with(flat_size, Default::default);
        Self {
            meta: SimMeta {
                dim: *size,
                inv_feature_size: R::one()/feature_size,
                time_step,
                feature_size,
            },
            e: vec![Vec3::zero(); flat_size],
            h: vec![Vec3::zero(); flat_size],
            m: vec![Vec3::zero(); flat_size],
            mat,
            step_no: 0,
            backend,
            diag: SyncDiagnostics::new(),
        }
    }
}

impl<R: Real, M: Default, B: Default> SpirvSim<R, M, B>
{
    pub fn new(size: Index, feature_size: f32) -> Self {
        Self::new_with_backend(size, feature_size, B::default())
    }
}

impl<R, M, B> SpirvSim<R, M, B> {
    fn flat_index(&self, idx: Index) -> Option<usize> {
        let dim = self.meta.dim;
        if idx.x() < dim.x() && idx.y() < dim.y() && idx.z() < dim.z() {
            Some(((idx.z()*dim.y() + idx.y())*dim.x() + idx.x()) as usize)
        } else {
            None
        }
    }
}

impl<R, M, B> SpirvSim<R, M, B>
where
    R: Real,
    M: Send + Sync + Material<R> + Clone + Into<FullyGenericMaterial<R>>,
    B: Send + Sync + SimBackend<R, M>,
{
    #[allow(unused)] // used for test
    fn apply_stimulus(&mut self, stim: &dyn AbstractStimulus) {
        trace!("apply_stimulus begin");
        iterate_stim(stim, self.size(), self.feature_size(), self.time(), self.meta.time_step, |pos_idx, value_e, value_h| {
            let flat_idx = self.flat_index(pos_idx).unwrap();
            self.e[flat_idx] += value_e.cast();
            self.h[flat_idx] += value_h.cast();
        });
        trace!("apply_stimulus end");
    }

    fn eval_stimulus<S: AbstractStimulus>(&self, stim: &S)
        -> (Array3<Vec3<R>>, Array3<Vec3<R>>)
    {
        trace!("eval_stimulus begin");
        let (e, h) = self.diag.instrument_stimuli(|| {
            let dim = self.size();
            let feature_size = self.feature_size();
            let t_sec = self.time();
            let timestep = self.meta.time_step;

            // TODO(perf): do this in one loop!
            let e = ndarray::Zip::from(ndarray::indices(
                [dim.z() as usize, dim.y() as usize, dim.x() as usize]
            )).par_map_collect(|(z, y, x)| {
                let pos_idx = Index::new(x as _, y as _, z as _);
                let pos_meters = pos_idx.to_meters(feature_size);
                let densities = stim.at(t_sec, pos_meters);
                densities.e.cast::<R>() * timestep
            });
            let h = ndarray::Zip::from(ndarray::indices(
                [dim.z() as usize, dim.y() as usize, dim.x() as usize]
            )).par_map_collect(|(z, y, x)| {
                let pos_idx = Index::new(x as _, y as _, z as _);
                let pos_meters = pos_idx.to_meters(feature_size);
                let densities = stim.at(t_sec, pos_meters);
                densities.h.cast::<R>() * timestep
            });
            (e, h)
        });
        trace!("eval_stimulus end");
        (e, h)
    }
}

fn iterate_stim<R, S, F>(
    stim: &S, dim: Index, feature_size: f32, t_sec: f32, timestep: R, mut f: F
)
where
    R: Real,
    S: AbstractStimulus + ?Sized,
    F: FnMut(Index, Vec3<R>, Vec3<R>),
{
    // TODO: parallelize
    for z in 0..dim.z() {
        for y in 0..dim.y() {
            for x in 0..dim.x() {
                let pos_idx = Index::new(x, y, z);
                let pos_meters = pos_idx.to_meters(feature_size);
                let densities = stim.at(t_sec, pos_meters);
                let (value_e, value_h) = (densities.e.cast::<R>() * timestep, densities.h.cast::<R>() * timestep);
                f(pos_idx, value_e, value_h)
            }
        }
    }
}

#[cfg(test)]
mod test {
    use super::*;
    use crate::real::{R32, ToFloat as _};
    use crate::sim::legacy::{self, SimState};
    use crate::mat::{self, AnisomorphicConductor};
    fn mean_magnitude_e<S: AbstractSim>(sim: &S) -> f32 {
        (sim.map_sum_enumerated(|_pos: Index, cell| {
            cell.e().mag().to_f64()
        })/((sim.width() * sim.height() * sim.depth()) as f64)).to_f32()
    }
    fn mean_magnitude_h<S: AbstractSim>(sim: &S) -> f32 {
        (sim.map_sum_enumerated(|_pos: Index, cell| {
            cell.h().mag().to_f64()
        })/((sim.width() * sim.height() * sim.depth()) as f64)).to_f32()
    }
    fn mean_magnitude_m<S: AbstractSim>(sim: &S) -> f32 {
        (sim.map_sum_enumerated(|_pos: Index, cell| {
            cell.m().mag().to_f64()
        })/((sim.width() * sim.height() * sim.depth()) as f64)).to_f32()
    }

    fn assert_simstate_eq<B: SimBackend<f32, FullyGenericMaterial<f32>> + Send + Sync>(real_state: &SpirvSim<f32, FullyGenericMaterial<f32>, B>, ref_state: &SimState) {
        let rel_threshold = 1e-2;
        let mean_e = mean_magnitude_e(ref_state);
        let mean_h = mean_magnitude_h(ref_state);
        let mean_m = mean_magnitude_m(ref_state);
        let dim = real_state.size();

        for z in 0..dim.z() {
            for y in 0..dim.y() {
                for x in 0..dim.x() {
                    let idx = Index::new(x, y, z);
                    let real_cell = real_state.sample(idx);
                    let ref_cell = ref_state.sample(idx);

                    let diff_e = real_cell.e().cast::<R32>() - ref_cell.e();
                    let diff_h = real_cell.h().cast::<R32>() - ref_cell.h();
                    let diff_m = real_cell.m().cast::<R32>() - ref_cell.m();

                    // if x == 1 && y == 1 && z == 0 {
                    // if ref_cell.m() != Vec3::zero() {
                    //     println!("e {} v.s. {}", real_cell.e(), ref_cell.e());
                    //     println!("h {} v.s. {}", real_cell.h(), ref_cell.h());
                    //     println!("m {} v.s. {}", real_cell.m(), ref_cell.m());
                    //     println!("diff_e: {}; diff_h: {}, diff_m: {}", diff_e, diff_h, diff_m);
                    // }
                    assert!(
                        diff_e.mag().to_f32() <= mean_e * rel_threshold,
                        "idx.e: {:?}. {:?} v.s. {:?}", idx, real_cell.e(), ref_cell.e()
                    );

                    assert!(
                        diff_h.mag().to_f32() <= mean_h * rel_threshold,
                        "idx.h: {:?}. {:?} v.s. {:?}", idx, real_cell.h(), ref_cell.h()
                    );

                    assert!(
                        diff_m.mag().to_f32() <= mean_m * rel_threshold || diff_m.mag() <= ref_cell.m().mag() * rel_threshold.cast::<R32>(),
                        "idx.m: {:?}. {:?} v.s. {:?}", idx, real_cell.m(), ref_cell.m()
                    );
                }
            }
        }
    }

    mod backend_agnostic {
        use super::*;
        use crate::stim::{NoopStimulus, RngStimulus, UniformStimulus};
        pub fn do_smoke_small<B: SimBackend<f32, FullyGenericMaterial<f32>> + Send + Sync + Default>() {
            let mut state = SpirvSim::new_with_backend(Index::new(8, 8, 8), 1e-3, B::default());
            state.step();
        }

        pub fn do_smoke_med_7bit<B: SimBackend<f32, FullyGenericMaterial<f32>> + Send + Sync + Default>() {
            let mut state = SpirvSim::new_with_backend(Index::new(124, 124, 124), 1e-3, B::default());
            state.step();
        }

        pub fn do_smoke_med128<B: SimBackend<f32, FullyGenericMaterial<f32>> + Send + Sync + Default>() {
            let mut state = SpirvSim::new_with_backend(Index::new(128, 128, 128), 1e-3, B::default());
            state.step();
        }

        pub fn do_smoke_med_23bit<B: SimBackend<f32, FullyGenericMaterial<f32>> + Send + Sync + Default>() {
            let mut state = SpirvSim::new_with_backend(Index::new(127, 256, 256), 1e-3, B::default());
            state.step();
        }

        pub fn do_smoke_med_0x800000_indexing<B: SimBackend<f32, FullyGenericMaterial<f32>> + Send + Sync + Default>() {
            let mut state = SpirvSim::new_with_backend(Index::new(128, 256, 256), 1e-3, B::default());
            state.step();
        }

        pub fn do_smoke_med_0x800000_address_space<B: SimBackend<f32, FullyGenericMaterial<f32>> + Send + Sync + Default>() {
            let mut state = SpirvSim::new_with_backend(Index::new(170, 256, 256), 1e-3, B::default());
            state.step();
        }

        pub fn do_smoke_large<B: SimBackend<f32, FullyGenericMaterial<f32>> + Send + Sync + Default>() {
            let mut state = SpirvSim::new_with_backend(Index::new(326, 252, 160), 1e-3, B::default());
            state.step();
        }

        pub fn do_smoke_not_multiple_of_4<B: SimBackend<f32, FullyGenericMaterial<f32>> + Send + Sync + Default>() {
            let mut state = SpirvSim::new_with_backend(Index::new(3, 2, 5), 1e-3, B::default());
            state.step();
        }

        fn test_same_explicit<B: SimBackend<f32, FullyGenericMaterial<f32>> + Send + Sync + Default>(
            mut ref_state: SimState,
            mut dut_state: SpirvSim<f32, FullyGenericMaterial<f32>, B>,
            step_iters: u64,
            steps_per_iter: u32
        ) {
            for _ in 0..step_iters {
                ref_state.step_multiple(steps_per_iter, &NoopStimulus);
                dut_state.step_multiple(steps_per_iter, &NoopStimulus);
                assert_simstate_eq(&dut_state, &ref_state);
            }
        }

        fn test_same<B: SimBackend<f32, FullyGenericMaterial<f32>> + Send + Sync + Default>(seed: u64, step_iters: u64, steps_per_iter: u32, size: Index) {
            let mut cpu_state = SimState::new(size, 1e-3);
            cpu_state.apply_stimulus(&RngStimulus::new(seed));

            let mut spirv_state = SpirvSim::new_with_backend(size, 1e-3, B::default());
            spirv_state.apply_stimulus(&RngStimulus::new(seed));

            test_same_explicit(cpu_state, spirv_state, step_iters, steps_per_iter);
        }

        pub fn do_same_no_step_no_stim<B: SimBackend<f32, FullyGenericMaterial<f32>> + Send + Sync + Default>() {
            let ref_state = SimState::new(Index::new(8, 8, 8), 1e-3);
            let dut_state = SpirvSim::new_with_backend(Index::new(8, 8, 8), 1e-3, B::default());
            test_same_explicit(ref_state, dut_state, 1, 1);
        }

        pub fn do_same_1_step<B: SimBackend<f32, FullyGenericMaterial<f32>> + Send + Sync + Default>() {
            test_same::<B>(0x1234, 1, 1, Index::new(4, 4, 4));
        }

        pub fn do_same_1000_step<B: SimBackend<f32, FullyGenericMaterial<f32>> + Send + Sync + Default>() {
            test_same::<B>(0x1234, 1000, 1, Index::new(4, 4, 4));
        }

        pub fn do_same_not_multiple_of_4<B: SimBackend<f32, FullyGenericMaterial<f32>> + Send + Sync + Default>() {
            test_same::<B>(0x1234, 100, 1, Index::new(3, 2, 5));
        }

        pub fn do_same_100_steps_larger<B: SimBackend<f32, FullyGenericMaterial<f32>> + Send + Sync + Default>() {
            test_same::<B>(0x1234, 100, 1, Index::new(24, 20, 44));
        }

        pub fn do_same_100_steps_of_10<B: SimBackend<f32, FullyGenericMaterial<f32>> + Send + Sync + Default>() {
            test_same::<B>(0x1234, 100, 10, Index::new(24, 20, 44));
        }

        fn test_same_conductor<B: SimBackend<f32, FullyGenericMaterial<f32>> + Send + Sync + Default>(
            seed: u64, step_iters: u64, steps_per_iter: u32, size: Index
        ) {
            use rand::{Rng as _, SeedableRng as _};
            let mut rng = rand::rngs::StdRng::seed_from_u64(seed);
            let mut ref_state = SimState::new(size, 1e-3);
            let mut dut_state = SpirvSim::new_with_backend(size, 1e-3, B::default());

            for z in 0..size.z() {
                for y in 0..size.y() {
                    for x in 0..size.x() {
                        let cond = Vec3::new(
                            rng.gen_range(0.0..1e6),
                            rng.gen_range(0.0..1e6),
                            rng.gen_range(0.0..1e6),
                        );
                        ref_state.put_material(Index::new(x, y, z), AnisomorphicConductor::new(cond.cast()));
                        dut_state.put_material(Index::new(x, y, z), AnisomorphicConductor::new(cond.cast()));
                    }
                }
            }

            ref_state.apply_stimulus(&RngStimulus::new(seed));
            dut_state.apply_stimulus(&RngStimulus::new(seed));

            test_same_explicit(ref_state, dut_state, step_iters, steps_per_iter);
        }

        pub fn do_conductor_1_step<B: SimBackend<f32, FullyGenericMaterial<f32>> + Send + Sync + Default>() {
            test_same_conductor::<B>(0x1234, 1, 1, Index::new(4, 4, 4));
        }

        pub fn do_conductor_many_steps_larger<B: SimBackend<f32, FullyGenericMaterial<f32>> + Send + Sync + Default>() {
            test_same_conductor::<B>(0x1234, 100, 10, Index::new(96, 16, 8));
        }

        fn test_same_mb_ferromagnet<B: SimBackend<f32, FullyGenericMaterial<f32>> + Send + Sync + Default>(
            seed: u64, step_iters: u64, steps_per_iter: u32, size: Index
        ) {
            use rand::{Rng as _, SeedableRng as _};
            let mut rng = rand::rngs::StdRng::seed_from_u64(seed);
            let mut ref_state = SimState::new(size, 1e-3);
            let mut dut_state = SpirvSim::new_with_backend(size, 1e-3, B::default());

            for z in 0..size.z() {
                for y in 0..size.y() {
                    for x in 0..size.x() {
                        let b_start = rng.gen_range(-1e-4..1e-4);
                        // if b_start is negative, we need the endpoint to be > |b_start|:
                        let b_range = (-2.0f32*b_start).max(0.0) + rng.gen_range(1e-5..1e-2);
                        // It would be weird for an increase in B to trigger a DECREASE in H.
                        // We have mu0(M+H) = B, so
                        // mu0(dM/dB = dH/dB) = 1
                        //   -> dH/dB = mu0^-1 - dM/dB > 0
                        let slope = rng.gen_range(0.0..0.9*f32::mu0_inv());
                        let b_end = b_start + b_range;
                        let m_max = b_range * slope;
                        ref_state.put_material(Index::new(x, y, z), legacy::mat::MBPgram::new(
                            b_start.cast(), b_end.cast(), m_max.cast()
                        ));
                        dut_state.put_material(Index::new(x, y, z), coremem_cross::mat::MBPgram::new(
                            b_start, b_end, m_max
                        ));
                    }
                }
            }

            // XXX H stimulus on the ferromagnet behaves differently on the CPU than GPU
            // the CPU implementation is the more suspect one here (it doesn't consider material
            // parameters).
            ref_state.apply_stimulus(&RngStimulus::new_e(seed));
            dut_state.apply_stimulus(&RngStimulus::new_e(seed));

            test_same_explicit(ref_state, dut_state, step_iters, steps_per_iter);
        }

        pub fn do_mb_ferromagnet_1_step<B: SimBackend<f32, FullyGenericMaterial<f32>> + Send + Sync + Default>() {
            test_same_mb_ferromagnet::<B>(0x1234, 1, 1, Index::new(4, 4, 4));
        }

        pub fn do_mb_ferromagnet_100_steps_larger<B: SimBackend<f32, FullyGenericMaterial<f32>> + Send + Sync + Default>() {
            test_same_mb_ferromagnet::<B>(0x1234, 10, 10, Index::new(96, 16, 8));
        }

        // XXX these tests probably failed because they were allowing negative mu_r values
        // #[test]
        // fn mb_ferromagnet_diff_repro() {
        //     let size = Index::new(3, 3, 3);
        //     let seed = 0;  // 0 => 28, 4 => 28
        //     let steps = 1000;
        //     let ref_mat = mat::MBPgram::new(0e-4, 9.02e-3, 1e4).into();
        //     let dut_mat = mat::MBPgram::new(0e-4, 9.02e-3, 1e4);
        //     let mut ref_state = SimState::new(size, 1e-3);
        //     let mut dut_state = SpirvSim::new(size, 1e-3);
        //     ref_state.put_material(Index::new(1, 1, 1), ref_mat);
        //     dut_state.put_material(Index::new(1, 1, 1), dut_mat);

        //     ref_state.apply_stimulus(&RngStimulus::new(seed));
        //     dut_state.apply_stimulus(&RngStimulus::new(seed));

        //     test_same_explicit(ref_state, dut_state, steps);
        // }

        // #[test]
        // fn mb_ferromagnet_diff_minimal_repro() {
        //     let size = Index::new(3, 3, 1);
        //     let seed = 0;  // 0 => 28, 4 => 28
        //     let steps = 1000;
        //     let ref_mat = mat::MBPgram::new(0e-4, 9.02e-3, 1e4).into();
        //     let dut_mat = mat::MBPgram::new(0e-4, 9.02e-3, 1e4);
        //     let mut ref_state = SimState::new(size, 1e-3);
        //     let mut dut_state = SpirvSim::new(size, 1e-3);
        //     ref_state.put_material(Index::new(1, 1, 0), ref_mat);
        //     dut_state.put_material(Index::new(1, 1, 0), dut_mat);

        //     ref_state.apply_stimulus(&RngStimulus::new(seed));
        //     dut_state.apply_stimulus(&RngStimulus::new(seed));

        //     test_same_explicit(ref_state, dut_state, steps);
        // }

        fn test_smoke_mh_ferromagnet<B: SimBackend<f32, FullyGenericMaterial<f32>> + Send + Sync + Default>(
            seed: u64, steps: u64, size: Index
        ) {
            // XXX This doesn't do anything, except make sure we don't crash!
            use rand::{Rng as _, SeedableRng as _};
            let mut rng = rand::rngs::StdRng::seed_from_u64(seed);
            let mut dut_state = SpirvSim::new_with_backend(size, 1e-3, B::default());

            for z in 0..size.z() {
                for y in 0..size.y() {
                    for x in 0..size.x() {
                        if rng.gen_range(0.0..1.0) < 0.5 {
                            let h_start = rng.gen_range(1.0..1000.0);
                            let mu_r = rng.gen_range(1.1..100000.0);
                            let m_max = rng.gen_range(0.0..1_000_000.0);
                            dut_state.put_material(Index::new(x, y, z), mat::MHPgram::new(h_start, mu_r, m_max));
                        } else if rng.gen_range(0.0..1.0) < 0.5 {
                            dut_state.put_material(Index::new(x, y, z), mat::IsomorphicConductor::new(rng.gen_range(0.0..1e7)));
                        }
                    }
                }
            }

            dut_state.apply_stimulus(&RngStimulus::new(seed));
            for step in 0..steps {
                println!("step {}", step);
                dut_state.step();
            }
        }

        pub fn do_mh_ferromagnet_smoke_1_step<B: SimBackend<f32, FullyGenericMaterial<f32>> + Send + Sync + Default>() {
            test_smoke_mh_ferromagnet::<B>(0x1234, 1, Index::new(4, 4, 4));
        }

        pub fn do_mh_ferromagnet_smoke_100_steps_larger<B: SimBackend<f32, FullyGenericMaterial<f32>> + Send + Sync + Default>() {
            test_smoke_mh_ferromagnet::<B>(0x1234, 100, Index::new(328, 252, 160));
        }

        pub fn do_step_multiple_with_stim<B: SimBackend<f32, FullyGenericMaterial<f32>> + Send + Sync + Default>() {
            let size = Index::new(4, 12, 8);
            let mut ref_state = SimState::new(size, 1e-3);
            let mut dut_state = SpirvSim::new_with_backend(size, 1e-3, B::default());
            let stim = UniformStimulus::new_e(Vec3::new(1.0e15, 2.0e15, -3.0e15));
            for _ in 0..5 {
                ref_state.step_multiple(100, &stim);
                dut_state.step_multiple(100, &stim);
                assert_simstate_eq(&dut_state, &ref_state);
            }
        }
    }

    macro_rules! test_backend {
        ($shortname:ident, $backend:ty) => {
            mod $shortname {
                use super::backend_agnostic::*;
                use super::*;

                #[test]
                fn test_smoke_small() {
                    do_smoke_small::<$backend>();
                }
                #[test]
                fn test_smoke_med_7bit() {
                    do_smoke_med_7bit::<$backend>();
                }
                #[test]
                fn smoke_med128() {
                    do_smoke_med128::<$backend>();
                }
                #[test]
                fn smoke_med_23bit() {
                    do_smoke_med_23bit::<$backend>();
                }
                #[test]
                fn smoke_med_0x800000_indexing() {
                    do_smoke_med_0x800000_indexing::<$backend>();
                }
                #[test]
                fn smoke_med_0x800000_address_space() {
                    do_smoke_med_0x800000_address_space::<$backend>();
                }
                #[test]
                fn smoke_large() {
                    do_smoke_large::<$backend>();
                }
                #[test]
                fn smoke_not_multiple_of_4() {
                    do_smoke_not_multiple_of_4::<$backend>();
                }
                #[test]
                fn same_no_step_no_stim() {
                    do_same_no_step_no_stim::<$backend>();
                }
                #[test]
                fn same_1_step() {
                    do_same_1_step::<$backend>();
                }
                #[test]
                fn same_1000_step() {
                    do_same_1000_step::<$backend>();
                }
                #[test]
                fn same_not_multiple_of_4() {
                    do_same_not_multiple_of_4::<$backend>();
                }
                #[test]
                fn same_100_steps_larger() {
                    do_same_100_steps_larger::<$backend>();
                }
                #[test]
                fn same_100_steps_of_10() {
                    do_same_100_steps_of_10::<$backend>();
                }
                #[test]
                fn conductor_1_step() {
                    do_conductor_1_step::<$backend>();
                }
                #[test]
                fn conductor_many_steps_larger() {
                    do_conductor_many_steps_larger::<$backend>();
                }
                #[test]
                fn mb_ferromagnet_1_step() {
                    do_mb_ferromagnet_1_step::<$backend>();
                }
                #[test]
                fn mb_ferromagnet_100_steps_larger() {
                    do_mb_ferromagnet_100_steps_larger::<$backend>();
                }
                #[test]
                fn mh_ferromagnet_smoke_1_step() {
                    do_mh_ferromagnet_smoke_1_step::<$backend>();
                }
                #[test]
                fn mh_ferromagnet_smoke_100_steps_larger() {
                    do_mh_ferromagnet_smoke_100_steps_larger::<$backend>();
                }
                #[test]
                fn step_multiple_with_stim() {
                    do_step_multiple_with_stim::<$backend>();
                }
            }
        }
    }

    test_backend!(cpu, CpuBackend);
    test_backend!(wgpu, WgpuBackend);
}
