use core::ops::{Index, IndexMut};

use crate::vec::Vec3u;

/// use this to wrap a flat region of memory into something which can be indexed by coordinates in
/// 3d space.
pub struct DimensionedSlice<T> {
    dim: Vec3u,
    items: T,
}

impl<T> DimensionedSlice<T> {
    pub fn new(dim: Vec3u, items: T) -> Self {
        Self { dim, items }
    }
}

impl<'a, T: Index<usize> + ?Sized> Index<Vec3u> for DimensionedSlice<&'a T> {
    type Output=T::Output;

    fn index(&self, idx: Vec3u) -> &Self::Output {
        let idx = index(idx, self.dim);
        &self.items[idx]
    }
}

impl<'a, T: Index<usize> + ?Sized> Index<Vec3u> for DimensionedSlice<&'a mut T> {
    type Output=T::Output;

    fn index(&self, idx: Vec3u) -> &Self::Output {
        let idx = index(idx, self.dim);
        &self.items[idx]
    }
}

impl<'a, T: IndexMut<usize> + ?Sized> IndexMut<Vec3u> for DimensionedSlice<&'a mut T> {
    fn index_mut(&mut self, idx: Vec3u) -> &mut Self::Output {
        let idx = index(idx, self.dim);
        &mut self.items[idx]
    }
}

fn index(loc: Vec3u, dim: Vec3u) -> usize {
    ((loc.z()*dim.y() + loc.y())*dim.x() + loc.x()) as usize
}


#[cfg(test)]
mod test {
    use super::*;
    #[test]
    fn test_index() {
        let dim = Vec3u::new(2, 3, 7);
        assert_eq!(index(Vec3u::new(0, 0, 0), dim), 0);
        assert_eq!(index(Vec3u::new(1, 0, 0), dim), 1);
        assert_eq!(index(Vec3u::new(0, 1, 0), dim), 2);
        assert_eq!(index(Vec3u::new(1, 1, 0), dim), 3);
        assert_eq!(index(Vec3u::new(0, 2, 0), dim), 4);
        assert_eq!(index(Vec3u::new(0, 0, 1), dim), 6);
        assert_eq!(index(Vec3u::new(1, 0, 1), dim), 7);
        assert_eq!(index(Vec3u::new(0, 1, 1), dim), 8);
        assert_eq!(index(Vec3u::new(1, 2, 1), dim), 11);
        assert_eq!(index(Vec3u::new(1, 2, 2), dim), 17);
    }
}

